if exist ..\..\clutter\clutter-enum-types.h goto DONE_COGL_ENUMS_H
perl %1\bin\glib-mkenums --template ../../clutter/clutter-enum-types.h.in ../../clutterclutter-action.h ../../clutterclutter-actor-meta.h ../../clutterclutter-actor.h ../../clutterclutter-align-constraint.h ../../clutterclutter-animatable.h ../../clutterclutter-backend.h ../../clutterclutter-bind-constraint.h ../../clutterclutter-binding-pool.h ../../clutterclutter-bin-layout.h ../../clutterclutter-blur-effect.h ../../clutterclutter-box-layout.h ../../clutterclutter-brightness-contrast-effect.h ../../clutterclutter-cairo.h ../../clutterclutter-canvas.h ../../clutterclutter-child-meta.h ../../clutterclutter-click-action.h ../../clutterclutter-cogl-compat.h ../../clutterclutter-clone.h ../../clutterclutter-color-static.h ../../clutterclutter-color.h ../../clutterclutter-colorize-effect.h ../../clutterclutter-constraint.h ../../clutterclutter-container.h ../../clutterclutter-content.h ../../clutterclutter-deform-effect.h ../../clutterclutter-deprecated.h ../../clutterclutter-desaturate-effect.h ../../clutterclutter-device-manager.h ../../clutterclutter-drag-action.h ../../clutterclutter-drop-action.h ../../clutterclutter-effect.h ../../clutterclutter-enums.h ../../clutterclutter-event.h ../../clutterclutter-feature.h ../../clutterclutter-fixed-layout.h ../../clutterclutter-flow-layout.h ../../clutterclutter-gesture-action.h ../../clutterclutter-grid-layout.h ../../clutterclutter-group.h ../../clutterclutter-image.h ../../clutterclutter-input-device.h ../../clutterclutter-interval.h ../../clutterclutter-keyframe-transition.h ../../clutterclutter-keysyms.h ../../clutterclutter-layout-manager.h ../../clutterclutter-layout-meta.h ../../clutterclutter-macros.h ../../clutterclutter-main.h ../../clutterclutter-offscreen-effect.h ../../clutterclutter-page-turn-effect.h ../../clutterclutter-paint-nodes.h ../../clutterclutter-paint-node.h ../../clutterclutter-pan-action.h ../../clutterclutter-path-constraint.h ../../clutterclutter-path.h ../../clutterclutter-property-transition.h ../../clutterclutter-rotate-action.h ../../clutterclutter-script.h ../../clutterclutter-scriptable.h ../../clutterclutter-scroll-actor.h ../../clutterclutter-settings.h ../../clutterclutter-shader-effect.h ../../clutterclutter-shader-types.h ../../clutterclutter-swipe-action.h ../../clutterclutter-snap-constraint.h ../../clutterclutter-stage.h ../../clutterclutter-stage-manager.h ../../clutterclutter-tap-action.h ../../clutterclutter-test-utils.h ../../clutterclutter-texture.h ../../clutterclutter-text.h ../../clutterclutter-text-buffer.h ../../clutterclutter-timeline.h ../../clutterclutter-transition-group.h ../../clutterclutter-transition.h ../../clutterclutter-types.h ../../clutterclutter-units.h ../../clutterclutter-zoom-action.h ../../clutterevdev/clutter-evdev.h ../../clutter/win32/clutter-win32.h > ..\..\clutter\clutter-enum-types.h
:DONE_COGL_ENUMS_H
if exist ..\..\clutter\clutter-enum-types.c goto DONE_COGL_ENUMS_C
perl %1\bin\glib-mkenums --template ../../clutter/clutter-enum-types.c.in ../../clutterclutter-action.h ../../clutterclutter-actor-meta.h ../../clutterclutter-actor.h ../../clutterclutter-align-constraint.h ../../clutterclutter-animatable.h ../../clutterclutter-backend.h ../../clutterclutter-bind-constraint.h ../../clutterclutter-binding-pool.h ../../clutterclutter-bin-layout.h ../../clutterclutter-blur-effect.h ../../clutterclutter-box-layout.h ../../clutterclutter-brightness-contrast-effect.h ../../clutterclutter-cairo.h ../../clutterclutter-canvas.h ../../clutterclutter-child-meta.h ../../clutterclutter-click-action.h ../../clutterclutter-cogl-compat.h ../../clutterclutter-clone.h ../../clutterclutter-color-static.h ../../clutterclutter-color.h ../../clutterclutter-colorize-effect.h ../../clutterclutter-constraint.h ../../clutterclutter-container.h ../../clutterclutter-content.h ../../clutterclutter-deform-effect.h ../../clutterclutter-deprecated.h ../../clutterclutter-desaturate-effect.h ../../clutterclutter-device-manager.h ../../clutterclutter-drag-action.h ../../clutterclutter-drop-action.h ../../clutterclutter-effect.h ../../clutterclutter-enums.h ../../clutterclutter-event.h ../../clutterclutter-feature.h ../../clutterclutter-fixed-layout.h ../../clutterclutter-flow-layout.h ../../clutterclutter-gesture-action.h ../../clutterclutter-grid-layout.h ../../clutterclutter-group.h ../../clutterclutter-image.h ../../clutterclutter-input-device.h ../../clutterclutter-interval.h ../../clutterclutter-keyframe-transition.h ../../clutterclutter-keysyms.h ../../clutterclutter-layout-manager.h ../../clutterclutter-layout-meta.h ../../clutterclutter-macros.h ../../clutterclutter-main.h ../../clutterclutter-offscreen-effect.h ../../clutterclutter-page-turn-effect.h ../../clutterclutter-paint-nodes.h ../../clutterclutter-paint-node.h ../../clutterclutter-pan-action.h ../../clutterclutter-path-constraint.h ../../clutterclutter-path.h ../../clutterclutter-property-transition.h ../../clutterclutter-rotate-action.h ../../clutterclutter-script.h ../../clutterclutter-scriptable.h ../../clutterclutter-scroll-actor.h ../../clutterclutter-settings.h ../../clutterclutter-shader-effect.h ../../clutterclutter-shader-types.h ../../clutterclutter-swipe-action.h ../../clutterclutter-snap-constraint.h ../../clutterclutter-stage.h ../../clutterclutter-stage-manager.h ../../clutterclutter-tap-action.h ../../clutterclutter-test-utils.h ../../clutterclutter-texture.h ../../clutterclutter-text.h ../../clutterclutter-text-buffer.h ../../clutterclutter-timeline.h ../../clutterclutter-transition-group.h ../../clutterclutter-transition.h ../../clutterclutter-types.h ../../clutterclutter-units.h ../../clutterclutter-zoom-action.h ../../clutterevdev/clutter-evdev.h ../../clutter/win32/clutter-win32.h > ..\..\clutter\clutter-enum-types.c
:DONE_COGL_ENUMS_C
